from ByteOperations import ByteOperations
from CRC8 import CRC8
import constants
from app_logging import AppLogger


class CommandGenerator:
    firstByte = 0xC5

    class FlagByte:
        NoFlag = 0x00
        Urgent = 0x02
        WaitingForQueued = 0x04

    class CommandByte:
        """
        Enum for command codes and size of their accompanying bytes. For example, the Form Network request is 0x20, and has 5 extra bytes of data, so the length is 5. If multiple sizes are possible, the largest one is listed.
        """
        FormNetworkRequest = (0x20, 5)
        UpdateNetworkStateRequest = (0x21, 5)
        RegisterWirelessDeviceRequest = (0x22, 4)
        NetworkStatusReport = (0x23, 10)
        QueuedMessageRequest = (0x24, 5)
        WirelessDeviceStatusReport = (0x25, 7)
        RegisteredDeviceListRequest = (0x26, 1)
        RegisteredSensorListResponse = (0x27, 17)
        ReadDataSectorRequest = (0x70, 5)
        ReadDataSectorResponse = (0x71, 38)
        WriteDataSectorRequest = (0x72, 37)
        WriteDataSectorResponse = (0x73, 6)
        ApplicationCommandRequest = (0x74, 5)  # CAN BE UP TO 43... so why isn't it?
        ApplicationCommandResponse = (0x75, 43)
        DataMessageReport = (0x55, 42)
        DataLoggedSensorDataMessage = (0x56, 46)
        ParentMessage = (0x52, 12)
        UpdateRapidPayload = (0xD9, 128)

    @staticmethod
    def getFormNetworkRequest(ChannelMask, NetworkID, Flag):
        log = AppLogger()
        assert type(ChannelMask) is list
        assert type(NetworkID) is int
        assert type(Flag) is int
        command = [CommandGenerator.firstByte]
        commandPair = CommandGenerator.CommandByte.FormNetworkRequest
        command.append(commandPair[1]+2)
        command.append(Flag)
        command.append(commandPair[0])
        command += ChannelMask
        command.append(NetworkID)
        command.append(CRC8.getCRC(command))
        if constants.VERBOSE_LOGS:
            log.info(constants.APP_NAME, 'Generated Command: {}'.format(command))
        return command

    @staticmethod
    def getQueuedMessageRequest(DeviceID, MessageStatus, Flag):
        log = AppLogger()
        assert type(DeviceID) is list
        assert type(MessageStatus) is int
        command = [CommandGenerator.firstByte]
        commandPair = CommandGenerator.CommandByte.QueuedMessageRequest
        command.append(commandPair[1]+2)
        command.append(Flag)
        command.append(commandPair[0])
        command += DeviceID
        command.append(MessageStatus)
        command.append(CRC8.getCRC(command))
        if constants.VERBOSE_LOGS:
            log.info('QueuedMessageRequest: : {}'.format(ByteOperations.bytToStr(command)))
        return command

    @staticmethod
    def addNetworkDeviceRequest(DeviceID, Flag):
        assert type(DeviceID) is list
        assert type(Flag) is int
        command = [CommandGenerator.firstByte]
        commandPair = CommandGenerator.CommandByte.RegisterWirelessDeviceRequest
        command.append(commandPair[1]+2)
        command.append(Flag)
        command.append(commandPair[0])
        command += DeviceID

        command.append(CRC8.getCRC(command))
        return command

    @staticmethod
    def getUpdateNetworkStateRequest(state, timearray, Flag):
        assert type(state) is int
        assert type(timearray) is list
        assert type(Flag) is int
        command = [CommandGenerator.firstByte]
        commandPair = CommandGenerator.CommandByte.UpdateNetworkStateRequest
        command.append(commandPair[1]+2)
        command.append(Flag)
        command.append(commandPair[0])
        command.append(state)
        command += timearray
        command.append(CRC8.getCRC(command))
        return command

    @staticmethod
    def getClearDownMessageQueueRequest(DeviceID, Flag):
        assert type(DeviceID) is list
        assert type(Flag) is int
        command = [CommandGenerator.firstByte]
        commandPair = CommandGenerator.CommandByte.ApplicationCommandRequest
        command.append(commandPair[1]+2)
        command.append(Flag)
        command.append(commandPair[0])
        command += DeviceID
        command.append(0x15)
        command.append(CRC8.getCRC(command))
        return command


    @staticmethod
    def getDownImageChunk(DeviceID, idx, payload):
        assert type(DeviceID) is list
        assert type(payload) is list
        command = [CommandGenerator.firstByte]
        commandPair = CommandGenerator.CommandByte.UpdateRapidPayload  # start delimiter
        command.append(128 + 2 + 6)  # length
        command.append(0)                   # options
        command.append(commandPair[0])      # command
        command += DeviceID
        command += idx
        command += payload
        # command.append(payload)             # data
        command.append(CRC8.getCRC(command))
        return command
