from datetime import datetime, timedelta

from ByteOperations import ByteOperations as BO
from SensorType import SENSOR_TYPES


class Field(object):
    def __init__(self, bytes):
        if type(bytes) is list:
            self.size = len(bytes)
            self.bytes = bytes
        else:
            self.bytes = [bytes]
            self.size = 1

    def get_value(self):
        return BO.from_bytes(self.bytes)

    def __str__(self):
        return BO.bytToStr(self.bytes)

    def __repr__(self):
        return str(self)

    def __len__(self):
        return self.size


class SensorTypeField(Field):
    def get_value(self):
        type_number = self.get_number()
        if type_number in SENSOR_TYPES:
            sensor_type = SENSOR_TYPES[type_number](self.bytes)
            return sensor_type.get_type()
        return str(type_number)

    def get_number(self):
        type_number = int.from_bytes(self.bytes, byteorder='little', signed=False)
        return type_number

    def __str__(self):
        return str(self.get_value())

    def __repr__(self):
        return str(self)


class TimeField(Field):

    def __init__(self, bytes, args):
        super().__init__(bytes)
        self.args = args

    def get_value(self):
        try:
            import pytz
            from tzlocal import get_localzone
            MONNIT_EPOCH = datetime(2010, 1, 1, tzinfo=pytz.utc)
            pytzinstalled = True
        except ImportError:
            MONNIT_EPOCH = datetime(2010, 1, 1)
            pytzinstalled = False
        seconds_from_epoch = int.from_bytes(self.bytes, byteorder='little', signed=False)

        time = MONNIT_EPOCH + timedelta(seconds=seconds_from_epoch)
        if self.args.use_local_time and pytzinstalled:
            return time.astimezone(get_localzone())
        else:
            return time

    def __str__(self):
        return str(self.get_value())


class VoltageField(Field):
    def get_value(self):
        return (150.0 + BO.from_bytes(self.bytes)) / 100

    def __str__(self):
        return str(self.get_value())


class SignedField(Field):
    def get_value(self):
        return int.from_bytes(self.bytes, byteorder='little', signed=True)

    def __str__(self):
        return str(self.get_value())


class BytesField(Field):
    def get_value(self):
        return self.bytes

    def __str__(self):
        return str(self.get_value())


class HexField(Field):
    def get_value(self):
        return hex(self.bytes[0])

    def __str__(self):
        return str(self.get_value())
