import decimal
import json
import pprint
import sys
from app_logging import AppLogger
import constants

from CRC8 import CRC8
from Field import BytesField, Field, HexField, SensorTypeField, SignedField, TimeField, VoltageField
from SensorType import SENSOR_TYPES


class PayloadMessage(object):

    def __init__(self, byt):
        if not CRC8.VerifyCRC(byt):
            raise Exception("Invalid CRC8 Checksum, retrying...")

        self.delimiter = HexField(byt[0])
        assert self.delimiter.bytes[0] == 0xC5
        self.length = Field(byt[1])
        self.options = Field(byt[2])
        self.command = HexField(byt[3])
        self.CRC8 = HexField(byt[-1])

    @staticmethod
    def json_encoder(obj):
        if isinstance(obj, HexField):
            return obj.get_value()
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        if isinstance(obj, SensorTypeField):
            return str(obj.get_value())
        if isinstance(obj, TimeField):
            return str(obj.get_value())
        if isinstance(obj, VoltageField):
            return float(obj.get_value())
        if isinstance(obj, BytesField):
            return obj.get_value()
        if isinstance(obj, Field):
            return int(obj.get_value())
        raise TypeError

    def as_json(self):
        return json.loads(str(self))

    def __str__(self):
        return json.dumps(self.__dict__, default=PayloadMessage.json_encoder)

    def write_to_std_out(self):
        log = AppLogger()
        if PayloadMessage.args.quiet:
            return
        # if constants. and self.command.bytes[0] != 0x56:
            # return
        if PayloadMessage.args.pretty:
            pp = pprint.PrettyPrinter(indent=4)
            pp.pprint(self.as_json())
            if constants.VERBOSE_LOGS:
                log.info('PayloadMessage: {}'.format(pprint.pformat(self.as_json(), indent=4)))
        else:
            if constants.VERBOSE_LOGS:
                log.info('PayloadMessage: {}'.format(self))
        if constants.datalog is True:
            bob = self.as_json()
            with open('datalog.txt', 'a') as f:
                # f.write(self.as_json())
                json.dump(bob, f, ensure_ascii=False)
                f.write('\n')
        sys.stdout.flush()


class DataLoggedSensorDataMessage(PayloadMessage):
    # Data Logged Sensor Data Message, 0x56
    def __init__(self, byt):
        super().__init__(byt)
        payload = byt[4:(4 + self.length.get_value() - 2)]

        self.device_id = Field(payload[0:4])
        self.time = TimeField(payload[4:8], self.args)
        self.field_rssi = SignedField(payload[8])
        self.ap_rssi = SignedField(payload[9])
        self.voltage = VoltageField(payload[10])
        self.sensor_type = SensorTypeField(payload[11:13])
        self.state = Field(payload[13])
        self.data = BytesField(payload[14:])

        if self.sensor_type.get_number() in SENSOR_TYPES.keys():
            sensor_type_class = SENSOR_TYPES[self.sensor_type.get_number()]
            datum_generator = sensor_type_class(self.data.bytes)
            self.datums = datum_generator.get_datums()


class ParentMessage(PayloadMessage):
    def __init__(self, byt):
        super().__init__(byt)
        payload = byt[4:(4 + self.length.get_value() - 2)]

        self.device_id = Field(payload[0:4])
        self.parent_id = Field(payload[5:8])
        self.version = Field(payload[8:12])


class WirelessDeviceStatusReport(PayloadMessage):
    def __init__(self, byt):
        super().__init__(byt)
        payload = byt[4:(4 + self.length.get_value() - 2)]

        self.device_id = Field(payload[0:4])
        self.node_type = BytesField(payload[4:6])
        self.status = Field(payload[6])


class WriteDataSectorResponse(PayloadMessage):
    def __init__(self, byt):
        super().__init__(byt)
        payload = byt[4:(4 + self.length.get_value() - 2)]

        self.device_id = Field(payload[0:4])
        self.sector_number = Field(payload[4])
        self.status = Field(payload[5])


class NetworkStatusReport(PayloadMessage):
    def __init__(self, byt):
        super().__init__(byt)
        payload = byt[4:(4 + self.length.get_value() - 2)]

        self.apn_serial_id = Field(payload[0:4])
        self.network_device_count = Field(payload[4:6])
        self.selected_channel = Field(payload[6])
        self.network_id = Field(payload[7])
        self.current_state = Field(payload[8])
        self.protocol_version = Field(payload[9])
        self.status = Field(payload[9])

PAYLOAD_MESSAGE_TYPE = {
    0x25: WirelessDeviceStatusReport,
    0x56: DataLoggedSensorDataMessage,
    #0x73: WriteDataSectorResponse,
    0x52: ParentMessage
}
