from abc import ABCMeta, abstractmethod


def bytes_to_temperature(temperature_bytes):
    temperature_int = int.from_bytes(temperature_bytes, byteorder='little', signed=True)
    temperature_float = (temperature_int/10.0)  # * 9.0/5.0 + 32  # Uncomment for fahrenheit
    return temperature_float


class SensorType(metaclass=ABCMeta):
    def __init__(self, bytearray):
        self.bytearray = bytearray

    @classmethod
    def get_type_index(cls):
        pass

    @classmethod
    def get_type(cls):
        pass

    @abstractmethod
    def get_datums(self):
        pass


class Temperature(SensorType):
    @classmethod
    def get_type_index(cls):
        return 2

    @classmethod
    def get_type(cls):
        return 'TEMPERATURE'

    def get_datums(self):
        temperature_bytes = self.bytearray[:2]
        return {'TEMPERATURE': bytes_to_temperature(temperature_bytes)}


class DryContact(SensorType):
    @classmethod
    def get_type_index(cls):
        return 3

    @classmethod
    def get_type(cls):
        return 'DRY_CONTACT'

    def get_datums(self):
        return {'CONTACT': self.bytearray[0] == 1}


class WaterDetect(SensorType):
    @classmethod
    def get_type_index(cls):
        return 4

    @classmethod
    def get_type(cls):
        return 'WATER_DETECT'

    def get_datums(self):
        return {'WATER': self.bytearray[0] == 1}


class Vibration(SensorType):
    @classmethod
    def get_type_index(cls):
        return 5

    @classmethod
    def get_type(cls):
        return 'VIBRATION'

    def get_datums(self):
        return {'VIBRATION': self.bytearray[0] == 1}


class MagneticPresence(SensorType):
    @classmethod
    def get_type_index(cls):
        return 6

    @classmethod
    def get_type(cls):
        return 'MAGNETIC_PRESENCE'

    def get_datums(self):
        return {'MAGNET PRESENT': self.bytearray[0] == 1}


class OpenClosed(SensorType):
    @classmethod
    def get_type_index(cls):
        return 9

    @classmethod
    def get_type(cls):
        return 'OPEN/CLOSED'

    def get_datums(self):
        return {'CLOSED': self.bytearray[0] == 1}


class Button(SensorType):
    @classmethod
    def get_type_index(cls):
        return 11

    @classmethod
    def get_type(cls):
        return 'BUTTON'

    def get_datums(self):
        return {'PRESSED': self.bytearray[0] == 1}


class Control(SensorType):
    @classmethod
    def get_type_index(cls):
        return 12

    @classmethod
    def get_type(cls):
        return 'CONTROL'

    def get_datums(self):
        relay1 = False
        relay2 = False
        if self.bytearray[0] & 0x01:
            relay1 = True
        if self.bytearray[0] & 0x02:
            relay2 = True
        return {'RELAY1': relay1, 'RELAY2': relay2}


class Snapshot(SensorType):
    @classmethod
    def get_type_index(cls):
        return 15

    @classmethod
    def get_type(cls):
        return 'ACCELSNAPSHOT'

    def get_datums(self):
        x = self.bytearray[0:2]
        y = self.bytearray[2:4]
        z = self.bytearray[4:6]
        x_int = int.from_bytes(x, byteorder='little', signed=True)
        y_int = int.from_bytes(y, byteorder='little', signed=True)
        z_int = int.from_bytes(z, byteorder='little', signed=True)
        x_float = x_int / 1000.0
        y_float = y_int / 1000.0
        z_float = z_int / 1000.0
        return {'X': x_float, 'Y': y_float, 'Z': z_float}


class ImpactDetect(SensorType):
    @classmethod
    def get_type_index(cls):
        return 16

    @classmethod
    def get_type(cls):
        return 'ACCELIMPACT'

    def get_datums(self):
        return {'IMPACT DETECTED': self.bytearray[0] >= 1}


class CurrentMetermA(SensorType):
    @classmethod
    def get_type_index(cls):
        return 22

    @classmethod
    def get_type(cls):
        return 'CURRENTMA'

    def get_datums(self):
        mA_bytes = self.bytearray[:2]
        mA_int = int.from_bytes(mA_bytes, byteorder='little', signed=False)
        mA_float = (mA_int / 100.0)
        return {'CURRENT': mA_float}


class VoltageMeter500VAC(SensorType):
    @classmethod
    def get_type_index(cls):
        return 32

    @classmethod
    def get_type(cls):
        return 'VOLTAGE_METER_500VDC'

    def get_datums(self):
        volt_bytes = self.bytearray[:2]
        volt_int = int.from_bytes(volt_bytes, byteorder='little', signed=False)
        volt_float = (volt_int / 10.0)
        return {'VOLTAGE': volt_float}


class TemperatureRTDHigh(SensorType):
    @classmethod
    def get_type_index(cls):
        return 35

    @classmethod
    def get_type(cls):
        return 'TEMPERATURE_RTD_HIGH'

    def get_datums(self):
        temperature_bytes = self.bytearray[:2]
        return {'TEMPERATURE': bytes_to_temperature(temperature_bytes)}


class HumidityTemperature(SensorType):
    @classmethod
    def get_type_index(cls):
        return 43

    @classmethod
    def get_type(cls):
        return 'HUMIDITY/TEMPERATURE'

    def get_datums(self):
        temp_bytes = self.bytearray[:2]
        temp_int = int.from_bytes(temp_bytes, byteorder='little', signed=False)
        temp_float = temp_int / 100.0
        humidity_bytes = self.bytearray[2:4]
        humidity_int = int.from_bytes(humidity_bytes, byteorder='little', signed=False)
        humidity_float = humidity_int / 100.0
        return {'HUMIDITY': humidity_float, 'TEMPERATURE': temp_float}


class TemperatureRTDLow(SensorType):
    @classmethod
    def get_type_index(cls):
        return 46

    @classmethod
    def get_type(cls):
        return 'TEMPERATURE_RTD_LOW'

    def get_datums(self):
        return {'TEMPERATURE': bytes_to_temperature(self.bytearray[:2])}


class VoltageMeter50VDC(SensorType):
    @classmethod
    def get_type_index(cls):
        return 59

    @classmethod
    def get_type(cls):
        return 'VOLTAGE_METER_50VDC'

    def get_datums(self):
        volt_bytes = self.bytearray[:2]
        volt_int = int.from_bytes(volt_bytes, byteorder='little', signed=False)
        volt_float = (volt_int / 1000.0)
        return {'VOLTAGE': volt_float}


class VoltageDetect500VAC(SensorType):
    @classmethod
    def get_type_index(cls):
        return 64

    @classmethod
    def get_type(cls):
        return 'VOLTAGE_DETECT_500VAC'

    def get_datums(self):
        return {'VOLTAGE PRESENT': self.bytearray[0] == 1}


class WaterTemperature(SensorType):
    @classmethod
    def get_type_index(cls):
        return 65

    @classmethod
    def get_type(cls):
        return 'WATER_TEMPERATURE'

    def get_datums(self):
        return {'TEMPERATURE': bytes_to_temperature(self.bytearray[:2])}


class UltrasonicCom(SensorType):
    @classmethod
    def get_type_index(cls):
        return 67

    @classmethod
    def get_type(cls):
        return 'COMMERCIAL_ULTRASONIC'

    def get_datums(self):
        dist_bytes = self.bytearray[:2]
        dist_int = int.from_bytes(dist_bytes, byteorder='little', signed=False)
        dist_float = dist_int
        return {'DISTANCE': dist_float}


class VoltageDetect50VDC(SensorType):
    @classmethod
    def get_type_index(cls):
        return 71

    @classmethod
    def get_type(cls):
        return 'VOLTAGE_DETECT_50VDC'

    def get_datums(self):
        return {'VOLTAGE PRESENT': self.bytearray[0] == 1}


class VoltageMeter5VDC(SensorType):
    @classmethod
    def get_type_index(cls):
        return 72

    @classmethod
    def get_type(cls):
        return 'VOLTAGE_METER_5VDC'

    def get_datums(self):
        volt_bytes = self.bytearray[:2]
        volt_int = int.from_bytes(volt_bytes, byteorder='little', signed=False)
        volt_float = (volt_int / 1000.0)
        return {'VOLTAGE': volt_float}


class VoltageMeter10VDC(SensorType):
    @classmethod
    def get_type_index(cls):
        return 74

    @classmethod
    def get_type(cls):
        return 'VOLTAGE_METER_10VDC'

    def get_datums(self):
        volt_bytes = self.bytearray[:2]
        volt_int = int.from_bytes(volt_bytes, byteorder='little', signed=False)
        volt_float = (volt_int / 1000.0)
        return {'VOLTAGE': volt_float}


class Tilt(SensorType):
    @classmethod
    def get_type_index(cls):
        return 75

    @classmethod
    def get_type(cls):
        return 'TILT'

    def get_datums(self):
        pitch = self.bytearray[0:2]
        roll = self.bytearray[2:4]
        pitch_int = int.from_bytes(pitch, byteorder='little', signed=True)
        roll_int = int.from_bytes(roll, byteorder='little', signed=True)
        # if pitch_int > 32768:
        #     pitch_int -= 65535
        pitch_float = pitch_int / 100.0
        # if roll_int > 32768:
        #     roll_int -= 65535
        roll_float = roll_int / 100.0
        return {'PITCH': pitch_float, 'ROLL': roll_float}


class WaterRope(SensorType):
    @classmethod
    def get_type_index(cls):
        return 78

    @classmethod
    def get_type(cls):
        return 'WATER_ROPE'

    def get_datums(self):
        return {'WATER': self.bytearray[0] == 1}


class Pressure50(SensorType):
    @classmethod
    def get_type_index(cls):
        return 79

    @classmethod
    def get_type(cls):
        return 'PRESSURE_50_PSI'

    def get_datums(self):
        pres_bytes = self.bytearray[:2]
        pres_int = int.from_bytes(pres_bytes, byteorder='little', signed=False)
        pres_float = (pres_int / 10.0)
        return {'PRESSURE': pres_float}


class Pressure300(SensorType):
    @classmethod
    def get_type_index(cls):
        return 81

    @classmethod
    def get_type(cls):
        return 'PRESSURE_300_PSI'

    def get_datums(self):
        pres_bytes = self.bytearray[:2]
        pres_int = int.from_bytes(pres_bytes, byteorder='little', signed=False)
        pres_float = (pres_int / 10.0)
        return {'PRESSURE': pres_float}


class PressureGen(SensorType):
    @classmethod
    def get_type_index(cls):
        return 82

    @classmethod
    def get_type(cls):
        return 'PRESSURE_CUSTOM'

    def get_datums(self):
        pres_bytes = self.bytearray[:2]
        pres_int = int.from_bytes(pres_bytes, byteorder='little', signed=False)
        pres_float = (pres_int / 10.0)
        return {'PRESSURE': pres_float}


class DuctTemperature(SensorType):
    @classmethod
    def get_type_index(cls):
        return 84

    @classmethod
    def get_type(cls):
        return 'DUCT_TEMPERATURE'

    def get_datums(self):
        return {'TEMPERATURE': bytes_to_temperature(self.bytearray[:2])}


class Thermocouple(SensorType):
    @classmethod
    def get_type_index(cls):
        return 86

    @classmethod
    def get_type(cls):
        return 'THERMOCOUPLE'

    def get_datums(self):
        temperature_bytes = self.bytearray[:2]
        return {'TEMPERATURE': bytes_to_temperature(temperature_bytes)}


class PulseCounter1Unit64Bit(SensorType):
    @classmethod
    def get_type_index(cls):
        return 90

    @classmethod
    def get_type(cls):
        return 'PULSE_COUNTER_1_UNIT_64_BIT'

    def get_datums(self):
        return {'COUNT': int.from_bytes(self.bytearray[:8], byteorder='little', signed=False)}


class QuadTemp(SensorType):
    @classmethod
    def get_type_index(cls):
        return 92

    @classmethod
    def get_type(cls):
        return 'QUAD_TEMP'

    def get_datums(self):
        probe1_bytes = self.bytearray[:2]
        probe2_bytes = self.bytearray[2:4]
        probe3_bytes = self.bytearray[4:6]
        probe4_bytes = self.bytearray[6:8]
        return {'PROBE1': bytes_to_temperature(probe1_bytes), 'PROBE2': bytes_to_temperature(probe2_bytes), 'PROBE3': bytes_to_temperature(probe3_bytes), 'PROBE4': bytes_to_temperature(probe4_bytes)}


class CurrentMeter20A(SensorType):
    @classmethod
    def get_type_index(cls):
        return 93

    @classmethod
    def get_type(cls):
        return 'CURRENT_METER_20A'

    def get_datums(self):
        amphours_bytes = int.from_bytes(self.bytearray[:8], byteorder='little', signed=False) / 100
        avgcurrent_bytes = int.from_bytes(self.bytearray[8:10], byteorder='little', signed=False) / 100
        maxcurrent_bytes = int.from_bytes(self.bytearray[10:12], byteorder='little', signed=False) / 100
        mincurrent_bytes = int.from_bytes(self.bytearray[12:14], byteorder='little', signed=False) / 100
        return {'AMPHOURS': amphours_bytes, 'AVERAGE CURRENT': avgcurrent_bytes, 'MAX CURRENT': maxcurrent_bytes, 'MIN CURRENT': mincurrent_bytes}


class CurrentMeter150A(SensorType):
    @classmethod
    def get_type_index(cls):
        return 94

    @classmethod
    def get_type(cls):
        return 'CURRENT_METER_150A'

    def get_datums(self):
        amphours_bytes = int.from_bytes(self.bytearray[:8], byteorder='little', signed=False) / 100
        avgcurrent_bytes = int.from_bytes(self.bytearray[8:10], byteorder='little', signed=False) / 100
        maxcurrent_bytes = int.from_bytes(self.bytearray[10:12], byteorder='little', signed=False) / 100
        mincurrent_bytes = int.from_bytes(self.bytearray[12:14], byteorder='little', signed=False) / 100
        return {'AMPHOURS': amphours_bytes, 'AVERAGE CURRENT': avgcurrent_bytes, 'MAX CURRENT': maxcurrent_bytes, 'MIN CURRENT': mincurrent_bytes}


class VibrationMeter200(SensorType):
    @classmethod
    def get_type_index(cls):
        return 95

    @classmethod
    def get_type(cls):
        return 'VIBRATIONMETER200'

    def get_datums(self):
        xvel = int(self.bytearray[0])
        yvel = int(self.bytearray[1])
        zvel = int(self.bytearray[2])
        xf = int(self.bytearray[3])
        yf = int(self.bytearray[4])
        zf = int(self.bytearray[5])
        xvel_float = xvel / 10.0
        yvel_float = yvel / 10.0
        zvel_float = zvel / 10.0
        dutyCycle = int(self.bytearray[6])
        return {'X-VELOCITY': xvel_float, 'Y-VELOCITY': yvel_float, 'Z-VELOCITY': zvel_float, 'X-FREQUENCY': xf, 'Y-FREQUENCY': yf, 'Z-FREQUENCY': zf, 'DUTYCYCLE': dutyCycle}


class FoodTemp(SensorType):
    @classmethod
    def get_type_index(cls):
        return 100

    @classmethod
    def get_type(cls):
        return 'FOOD_GRADE_TEMPERATURE'

    def get_datums(self):
        temperature_bytes = self.bytearray[:2]
        return {'TEMPERATURE': bytes_to_temperature(temperature_bytes)}


class PIRMotionDetect(SensorType):
    @classmethod
    def get_type_index(cls):
        return 101

    @classmethod
    def get_type(cls):
        return 'PIR_MOTION_DETECT'

    def get_datums(self):
        return {'MOTION_DETECTED': self.bytearray[0] == 1}


class PM25(SensorType):
    @classmethod
    def get_type_index(cls):
        return 102

    @classmethod
    def get_type(cls):
        return 'PM_25'

    def get_datums(self):
        pm1_int = int.from_bytes(self.bytearray[:2], byteorder='little', signed=False)
        pm25_int = int.from_bytes(self.bytearray[2:4], byteorder='little', signed=False)
        pm10_int = int.from_bytes(self.bytearray[4:6], byteorder='little', signed=False)
        return {'PM1': pm1_int, 'PM25': pm25_int, 'PM10': pm10_int}


class DifferentialPressure(SensorType):
    @classmethod
    def get_type_index(cls):
        return 103

    @classmethod
    def get_type(cls):
        return 'DIFFERENTIAL_PRESSURE/TEMPERATURE'

    def get_datums(self):
        pres_bytes = self.bytearray[:2]
        pres_int = int.from_bytes(pres_bytes, byteorder='little', signed=False)
        pres_float = pres_int/10.0
        temperature_bytes = self.bytearray[2:]
        return {'PRESSURE': pres_float, 'TEMPERATURE': bytes_to_temperature(temperature_bytes)}


class UltrasonicInd(SensorType):
    @classmethod
    def get_type_index(cls):
        return 105

    @classmethod
    def get_type(cls):
        return 'INDUSTRIAL_ULTRASONIC'

    def get_datums(self):
        dist_bytes = self.bytearray[:2]
        dist_int = int.from_bytes(dist_bytes, byteorder='little', signed=False)
        # dist_float = (dist_int / 10.0)
        return {'DISTANCE': dist_int}


class LightMeter(SensorType):
    @classmethod
    def get_type_index(cls):
        return 107

    @classmethod
    def get_type(cls):
        return 'LIGHT_METER'

    def get_datums(self):
        lux_bytes = self.bytearray[:4]
        lux_int = int.from_bytes(lux_bytes, byteorder='little', signed=False)
        lux_float = lux_int / 100.0
        return {'LUX': lux_float, 'LIGHT PRESENT': self.bytearray[4] == 1}


class VoltageMeter200VDC(SensorType):
    @classmethod
    def get_type_index(cls):
        return 113

    @classmethod
    def get_type(cls):
        return 'VOLTAGE_METER_200VDC'

    def get_datums(self):
        volt_bytes = self.bytearray[:2]
        volt_int = int.from_bytes(volt_bytes, byteorder='little', signed=False)
        volt_float = (volt_int / 100.0)
        return {'VOLTAGE': volt_float}


class AirVelocity(SensorType):
    @classmethod
    def get_type_index(cls):
        return 114

    @classmethod
    def get_type(cls):
        return 'AIR_VELOCITY/TEMPERATURE'

    def get_datums(self):
        speed_bytes = self.bytearray[2:]
        speed_int = int.from_bytes(speed_bytes, byteorder='little', signed=False)
        speed_float = speed_int/10.0
        temperature_bytes = self.bytearray[:2]
        return {'SPEED': speed_float, 'TEMPERATURE': bytes_to_temperature(temperature_bytes)}


SENSOR_TYPES = {
    2: Temperature,
    3: DryContact,
    4: WaterDetect,
    5: Vibration,
    9: OpenClosed,
    11: Button,
    12: Control,
    15: Snapshot,
    16: ImpactDetect,
    # 19: Activity2,
    22: CurrentMetermA,
    # 23: PIRMotionDetect,
    # 27: LightDetect,
    32: VoltageMeter500VAC,
    35: TemperatureRTDHigh,
    # 42: LoggedActivityTimer,
    43: HumidityTemperature,
    # 45: RangeExtender,
    46: TemperatureRTDLow,
    # 48: PulseCounter1Unit32Bit,
    # 51: SeatOccupancy,
    59: VoltageMeter50VDC,
    64: VoltageDetect500VAC,
    65: WaterTemperature,
    67: UltrasonicCom,
    71: VoltageDetect50VDC,
    72: VoltageMeter5VDC,
    74: VoltageMeter10VDC,
    75: Tilt,
    78: WaterRope,
    79: Pressure50,
    82: Pressure300,
    83: PressureGen,
    84: DuctTemperature,
    86: Thermocouple,
    90: PulseCounter1Unit64Bit,
    92: QuadTemp,
    93: CurrentMeter20A,
    94: CurrentMeter150A,
    95: VibrationMeter200,
    100: FoodTemp,
    101: PIRMotionDetect,
    102: PM25,
    103: DifferentialPressure,
    105: UltrasonicInd,
    107: LightMeter,
    113: VoltageMeter200VDC,
    114: AirVelocity
}
