#!/usr/bin/python3
import sys
import time
import constants
import threading
import os
import serial
import traceback

from confighandler import rebuild
from confighandler import lock_config_file
from confighandler import unlock_config_file
from configparser import ConfigParser
from ByteOperations import ByteOperations
from CommandGenerator import CommandGenerator as CmdGen
from DownMessageParser import DownMessageParser
from GatewayMessage import GatewayMessage
from GetUSBLocation import get_time
from MonnitTcp import TcpClient as TcpClient
from PayloadMessage import NetworkStatusReport, PayloadMessage
from SensorListRequest import SensorListRequest
from SerialCommunication import send_message
from SerialCommunication import get_message
from StartupMessage import startup_sequence
from arguments import get_parser
from encryption import encrypt_packet, decrypt_packet
from errors import USBException
from start_server import download_sensors
from app_logging import AppLogger
from DLImageRequest import DLImageRequest


arguments = []
error_retry_count = 0


# def add_sensors_from_server(ser, gwid, args):
#     sensor_list = SensorListRequest.GetAllSensors(gwid, args)
#     down_messages = set()
#     x = []
#     for sensor in sensor_list:
#         down_message = bytes(CmdGen.addNetworkDeviceRequest(ByteOperations.to_bytes(sensor, 4), 0x00))
#         x.append(down_message)
#     down_messages.update(x)
#     if down_messages:
#         down_message_handler(ser, args, down_messages)


def add_sensors_from_server(ser, gwid, args):
    sensor_list = SensorListRequest.GetAllSensors(gwid, args)
    down_messages = []
    x = []
    for sensor in sensor_list:
        down_message = bytes(CmdGen.addNetworkDeviceRequest(ByteOperations.to_bytes(sensor, 4), 0x00))
        down_messages.append(down_message)
    if down_messages:
        down_message_handler(ser, args, down_messages)


def down_message_handler(ser, args, down_messages):
    log = AppLogger()
    queue_clear_request = CmdGen.getClearDownMessageQueueRequest(args.gwid, 0x00)
    retry = 0
    retry_break = False
    response = []
    while retry < 3:
        send_message(ser, queue_clear_request)
        cnt = 0
        while cnt < 5:
            queue_clear_response = get_message(ser)
            cnt += 1
            try:
                if constants.VERBOSE_LOGS:
                    log.info('Down Queue clear request Response: {}'.format(queue_clear_response))  # protect print
                if queue_clear_response[3] == 0x75 and queue_clear_response[8] == 21:
                    retry_break = True
                    break
                else:
                    time.sleep(0.1)
            except Exception as dqcr:
                if queue_clear_response is not None:
                    log.warning('Down queue clear Requests Error: {}'.format(dqcr))
                continue
        if retry_break is True:
            break
        retry += 1
    for down_message in down_messages:
        int_response = [b for b in down_message[0:15]]
        retry = 0
        retry_break = False
        cnt = 0
        while retry < 3:
            send_message(ser, down_message)
            while cnt < 5:
                dm_response = get_message(ser)
                cnt += 1
                try:
                    if dm_response[3] == 0xD5 or dm_response[3] == 0xF5:
                        constants.BLFlag = True
                        constants.DeviceID = [dm_response[4], dm_response[5], dm_response[6], dm_response[7]]
                        constants.FWID = [dm_response[8], dm_response[9], dm_response[10], dm_response[11]]
                        constants.uLength = [dm_response[12], dm_response[13]]
                        constants.uFlags = [dm_response[14]]  # ByteOperations.to_bytes(15, 1)
                        constants.OTAType = [dm_response[3]]  # Designator of whether it is D5 or F5
                        if dm_response[3] == 0xF5:
                            constants.response = [z for z in dm_response]
                        else:
                            constants.sen_update = True
                    if down_message[3] == 0xD4 or down_message[3] == 0xF4:
                        # log.info('Waiting to see if message ever comes')
                        continue
                    elif dm_response[3] == 0xD8:  # is this all I need here?
                        retry_break = True
                        if down_message[8] == 0xFF and down_message[9] == 0xFF:
                            time.sleep(.1)
                        break
                    elif dm_response[3] == 0x25 and dm_response[10] == 0x02:
                        retry_break = True
                        break
                except Exception as dqmr:
                    if dm_response is None:
                        break
                    else:
                        log.warning('Down message Error: {}'.format(dqmr))
                        continue
            if retry_break is True:
                retry_break = False
                retry = 0
                break
            else:
                time.sleep(0.1)
            retry += 1


def deliver_messages_to_server(message_queue, urgent, args):
    if constants.server_com is True:
        log = AppLogger()
        try:
            down_messages = set()
            for i, messagePacket in enumerate(message_queue):
                is_last = i != (len(message_queue) - 1)
                use_more_flag = urgent or is_last
                up_message = messagePacket.getMessage(use_more_flag)
                up_message = encrypt_packet(up_message, args)
                if constants.VERBOSE_LOGS:
                    log.info('Sent to Monnit server: {}'.format(up_message))
                encrypted_response = TcpClient.send_data(bytes(up_message), arguments)
                response = decrypt_packet(encrypted_response, args)
                if constants.VERBOSE_LOGS:
                    log.info('Response Monnit server: {}'.format(response))
                additional_down_messages = DownMessageParser.process(response)
                down_messages.update(additional_down_messages)
            message_queue = []
            # If urgent, don't pass down messages
            if urgent:
                return message_queue, set()
            return message_queue, down_messages
        except ConnectionRefusedError as cre:
            usb_exception = USBException(str(cre) + '\nSaving message list for retry.\n').as_json()
            log.error('{}'.format(usb_exception))
            return message_queue, []
    else:
        return [], []


def send_up_messages(ser, up_message_queue, is_urgent, args):
    up_message_queue, down_messages = deliver_messages_to_server(up_message_queue, is_urgent, args)
    if down_messages:
        down_message_handler(ser, args, down_messages)
    return up_message_queue


def add_sensors(ser, sensor_list):
    for sensor in sensor_list:
        add_sensor_request = CmdGen.addNetworkDeviceRequest(ByteOperations.to_bytes(sensor, 4), 0x00)
        send_message(ser, add_sensor_request)
        # result = get_message(ser)


def activate_gateway(args, ser):
    log = AppLogger()
    utc_now = int(get_time().total_seconds())
    time_array = ByteOperations.to_bytes(utc_now, 4)
    active_message = CmdGen.getUpdateNetworkStateRequest(1, time_array, 0x00)
    retry = 0
    while True:
        send_message(ser, active_message)
        active_response = get_message(ser)
        # if active_response is None:
        #     log.info('The Activate gateway message was none for the {}'.format(retry))
        if active_response is not None:
            if active_response[3] == 0x23:
                break
        elif retry > 4:
            break
        retry += 1

    if len(active_response) == 0:
        usb_exception = USBException('No response, USB Pro may be reforming network.').as_json()
        log.error('{}'.format(usb_exception))
        raise Exception('No response, USB Pro may be reforming network.')
        # sys.exit()

    network_status_report = NetworkStatusReport(active_response)
    network_status_report.write_to_std_out()

    device_gwid_bytes = [int(j) for j in network_status_report.apn_serial_id.bytes]
    device_gwid = int.from_bytes(device_gwid_bytes, byteorder='little', signed=False)

    config = ConfigParser()
    try:
        lock_config_file('activate_gateway()')
        config.read('config.ini')
        config.set('usb_configs', 'state', 'Active')
        if not args.gwid:
            args.gwid = device_gwid_bytes
            GatewayMessage.GWID = args.gwid
            PayloadMessage.args = args
            config.set('usb_configs', 'gwid', str(device_gwid))
            with open('config.ini', 'w') as f:
                config.write(f)
    except Exception as ex:
        log.error('File read/write error in activate_gateway: {}'.format(ex))
    finally:
        unlock_config_file('activate_gateway()')
    return args


def alternate_run_modes(args, ser):
    log = AppLogger()
    config = ConfigParser()
    try:
        lock_config_file('alternate_run_modes()')
        config.read('config.ini')
        reform = config.getint('usb_configs', 'reform')
        download = config.getint('usb_configs', 'download')
        if download:
            log.info('Downloading new sensors.')
            if constants.server_com is True:
                add_sensors_from_server(ser, args.gwid, args)
            config.read('config.ini')
            config.set('usb_configs', 'download', '0')
            with open('config.ini', 'w') as f:
                config.write(f)

        if args.add_sensors:
            try:
                sensorList = [int(s) for s in args.add_sensors.split(',')]
                add_sensors(ser, sensorList)
            except ValueError:
                usb_exception = USBException('Sensors must be a comma separated list of int, ie 1,2,3').as_json()
                log.error('{}'.format(usb_exception))
        if reform:
            log.info('Reforming USB Pro')
            # Create a reform network request with all channels available and a random network id
            reform_network_request = CmdGen.getFormNetworkRequest([0xff, 0xff, 0xff, 0xff], 0x00, 0x00)
            send_message(ser, reform_network_request)
            response = get_message(ser)
            config.read('config.ini')
            config.set('usb_configs', 'reform', '0')
            with open('config.ini', 'w') as f:
                config.write(f)
    except Exception as ex:
        log.error('File read/write error in alternate_run_modes: {}'.format(ex))
    finally:
        unlock_config_file('alternate_run_modes()')


def get_messages_from_gateway(args, ser):
    log = AppLogger()
    message_dict = {}
    message_iteration = 0
    flush = True
    # Retrieve Message Loop
    while True:
        if flush is True:
            data = get_message(ser)
            if data is None:
                flush = False
                continue
        else:
            queued_message_request = CmdGen.getQueuedMessageRequest(args.gwid, message_iteration, 0x00)
            # log.info('Queued Message request sent down: {}'.format(queued_message_request))
            send_message(ser, queued_message_request)
            data = get_message(ser)
        if data is None:
            break
        data_response_len = data[1]
        data_response_cmd = data[3]
        data_response_code = data[8]
        device_id = [data[4], data[5], data[6], data[7]]
        if data_response_len != (len(data) - 3):
            log.warning('Message length mismatch, potential loss of data. {}'.format(len(data)))
        if data_response_cmd == 0x23:
            data_response_device_count = [data[8], data[9]]
            sensorcount = int.from_bytes(data_response_device_count, byteorder='little', signed=False)
            config = ConfigParser()
            try:
                lock_config_file('get_messages_from_gateway() - 1')
                config.read('config.ini')
                config.set('usb_configs', 'sensors', '{}'.format(sensorcount))
                with open('config.ini', 'w') as f:
                    config.write(f)
            except Exception as ex:
                log.error('File read/write error in msg 0x23: {}'.format(ex))
            finally:
                unlock_config_file('get_messages_from_gateway() - 1')
        elif data_response_cmd == 0x24 and data_response_code == 0x0C:
            if flush:
                continue
            else:
                break
        elif data_response_cmd == 0x24 and data_response_code == 0x0F:  # what would be the proper handling for this?
            log.info('0X24 MESSAGE with a wait- {}'.format(data_response_code))
            break
        elif data_response_cmd == 0x25 or data_response_cmd == 0x71 or data_response_cmd == 0x73 or data_response_cmd == 0x75:
            if device_id == args.gwid or data_response_code == 0x02:  # if gwid is same or received down message properly queued
                # device_id = None
                if constants.VERBOSE_LOGS is True:
                    if data[7] == 0x00:
                        print('Join Success: {}'.format(int.from_bytes(device_id, byteorder='little', signed=False)))
            else:
                message_time = get_time()
                message_dict[message_time] = data
                message_iteration += 1
                message_iteration %= 256
                if constants.VERBOSE_LOGS is True:
                    if data[7] == 0x00:
                        print('Join Success: {}'.format(int.from_bytes(device_id, byteorder='little', signed=False)))
        elif data_response_cmd == 0x56:
            message_time = get_time()
            message_dict[message_time] = data
            message_iteration += 1
            message_iteration %= 256
        elif data_response_cmd == 0xD5:
            if constants.BLFlag is False:
                constants.BLFlag = True
                constants.DeviceID = [data[4], data[5], data[6], data[7]]
                constants.FWID = [data[8], data[9], data[10], data[11]]
                constants.uLength = [data[12], data[13]]
                constants.uFlags = [data[14]]
                constants.OTAType = [data[3]]
                constants.sen_update = True
                message_time = get_time()
                message_dict[message_time] = data
                message_iteration += 1
                message_iteration %= 256
                config = ConfigParser()
                try:
                    lock_config_file('get_messages_from_gateway() - 2')
                    config.read('config.ini')
                    config.set('usb_configs', 'state', 'Busy')
                    config.set('usb_configs', 'send', '1')
                    with open('config.ini', 'w') as f:
                        config.write(f)
                except Exception as ex:
                    log.error('File read/write error in msg 0xD5: {}'.format(ex))
                finally:
                    unlock_config_file('get_messages_from_gateway() - 2')
        elif data_response_cmd == 0xD6 or data_response_cmd == 0xD8:
            break
        else:
            message_time = get_time()
            message_dict[message_time] = data
            message_iteration += 1
            message_iteration %= 256
    return message_dict


def kill_usb_pro():
    unlock_config_file('kill_usb_pro()')
    sys.exit(0)


def run_usb_pro(args, ser):
    log = AppLogger()
    global error_retry_count
    log.info('run_usb_pro() - Enter')

    # ser = None
    # for port2try in args.input:
    #     try:
    #         ser = serial.Serial(port=port2try, baudrate=115200, parity=serial.PARITY_NONE, timeout=.1)
    #         time.sleep(.1)
    #         args = activate_gateway(args, ser)
    #         break
    #     except IndexError:
    #         ser = None
    #         continue
    #
    # if ser is None:
    #     log.error('No USB Sensor Adapter Detected!')
    #     raise Exception('No USB Sensor Adapter Detected!')
    # else:
    #     log.info('Serial port: {}'.format(ser.port))

    # Check to make sure the server is accessible
    if not args.no_network:
        if not TcpClient.check_server_status(args):
            ser.close()
            kill_usb_pro()

    # Initialize Encryption options with server
    if args.encryption == '1':
        log.info('Encryption Enabled')
        args = startup_sequence(args)

    alternate_run_modes(args, ser)

    last_up_message_send_time = time.time()
    last_download_time = time.time()
    up_message_queue = []
    caused_by_queue_size = False

    MasterMessage_Dict = {}
    down_messages = []
    aware = 0
    send = 0
    while True:  # This runs every sleep time
        foundurgent = False
        if not constants.BLFlag:
            try:
                message_dict = get_messages_from_gateway(args, ser)
            except IndexError:
                continue

            lock_config_file('run_usb_pro() - 1')
            try:
                config = ConfigParser()
                config.read('config.ini')
                state = config.getint('state', 'usb')
                aware = config.getint('usb_configs', 'aware')
            except Exception as ex:
                log.error('File read/write error in run_usb_pro_1: {}'.format(ex))

            if len(message_dict) > 0:
                for time_key, message_value in message_dict.items():
                    MasterMessage_Dict[time_key] = message_value
                    foundurgent = foundurgent or ((message_value[2] & 0x2) != 0)

            if state == 0:
                log.info('Shutting down USB Pro.')
                msg_pack = GatewayMessage(MasterMessage_Dict, args)
                up_message_queue.append(msg_pack)
                send_up_messages(ser, up_message_queue, foundurgent, args)
                ser.close()
                kill_usb_pro()
            elif state == 2:
                msg_pack = GatewayMessage(MasterMessage_Dict, args)
                up_message_queue.append(msg_pack)
                send_up_messages(ser, up_message_queue, foundurgent, args)
                log.info('Restarting USB Pro.')
                config.set('state', 'usb', '1')
                with open('config.ini', 'w') as f:
                    config.write(f)

                ser.close()
                kill_usb_pro()
            unlock_config_file('run_usb_pro() - 1')

            # region Send Messages to Server
            elapsed_time = (time.time() - last_up_message_send_time) / 60.0  # time in minutes
            server_time_elapsed = elapsed_time > float(args.heartbeat)
            queue_size_reached = len(up_message_queue) > int(args.max_up_queue_size)
            try:
                send = config.getint('usb_configs', 'send')
            except Exception as ex:
                log.error('File read/write error in run_usb_pro_2: {}'.format(ex))
            # If messages were sent successfully, the queue will be reset, and it is okay to trigger on queue size
            if not queue_size_reached:
                caused_by_queue_size = False
            if server_time_elapsed or (queue_size_reached and not caused_by_queue_size) or (foundurgent and aware) or send:
                activate_gateway(args, ser)
                last_up_message_send_time = time.time()
                msgpack = GatewayMessage(MasterMessage_Dict, args)
                up_message_queue.append(msgpack)
                log.info('{} messages sent to the server'.format(len(up_message_queue)))
                up_message_queue, down_messages = deliver_messages_to_server(up_message_queue, foundurgent, args)
                caused_by_queue_size = queue_size_reached
                MasterMessage_Dict = {}
                lock_config_file('run_usb_pro() - 2')
                try:
                    config.read('config.ini')
                    config.set('usb_configs', 'send', '0')
                    with open('config.ini', 'w') as f:
                        config.write(f)
                except Exception as ex:
                    log.error('File read/write error in run_usb_pro_3: {}'.format(ex))
                unlock_config_file('run_usb_pro() - 2')
                if down_messages:
                    down_message_handler(ser, args, down_messages)
            try:
                lock_config_file('run_usb_pro() - 3')
                config.read('config.ini')
                config.set('usb_configs', 'state', 'Active')
                with open('config.ini', 'w') as f:
                    config.write(f)
            except Exception as ex:
                log.error('File read/write error in run_usb_pro_4: {}'.format(ex))
            finally:
                unlock_config_file('run_usb_pro() - 3')

            error_retry_count = 0

            time_since_download = (time.time() - last_download_time) / 60.0  # time in minutes
            download = time_since_download > (12 * 60)

            if download:
                download_sensors()
                last_download_time = time.time()
            # log.info(args.sleep_time)
            # time.sleep(int(args.sleep_time))
            time.sleep(0.5)
        else:
            log.info('Entering update process.')
            down_messages = []
            DLImageRequest.GetCompleteImage(args.gwid, args)
            FWlength = ByteOperations.from_bytes(constants.uLength)
            updatepos = 0
            endflag = False
            while True:
                with open('updateImage.txt', 'rb') as f:
                    f.seek((updatepos * 128), 0)
                    bob = updatepos * 128
                    if (bob + 128) < FWlength:
                        image = f.read(128)
                    else:
                        image = f.read(FWlength - updatepos)
                        endflag = True
                payload = [b for b in image]
                while len(payload) != 128:
                    payload.append(0xFF)
                upper = int(updatepos / 256)
                lower = updatepos % 256
                uidx = [lower, upper]
                # try:
                down_message = bytes(CmdGen.getDownImageChunk(constants.DeviceID, uidx, payload))
                # except Exception as W:
                down_messages.append(down_message)
                updatepos += 1
                if endflag is True:
                    break
            uidx = [0xFF, 0xFF]
            final_message = bytes(CmdGen.getDownImageChunk(constants.DeviceID, uidx, payload))
            down_messages.append(final_message)
            if down_messages:
                up_messages = []
                down_message_handler(ser, args, down_messages)
                if not constants.sen_update:
                    updict = {get_time(): constants.response}
                    msgpack = GatewayMessage(updict, args)
                    up_message_queue.append(msgpack)
                    deliver_messages_to_server(up_messages, True, args)
                    # if there are responses that need to go to the server push them up
                    time.sleep(5)
                    config = ConfigParser()
                    try:
                        lock_config_file('run_usb_pro() - 4')
                        config.read('config.ini')
                        config.set('usb_configs', 'state', 'Busy')  # may want reform
                        config.set('state', 'usb', '2')
                        with open('config.ini', 'w') as f:
                            config.write(f)
                    except Exception as ex:
                        log.error('File read/write error in run_usb_pro_5: {}'.format(ex))
                    finally:
                        unlock_config_file('run_usb_pro() - 4')
                constants.BLFlag = False

        time.sleep(1)


def serial_search(args, ser):
    log = AppLogger()
    for port2try in args.input:
        try:
            ser = serial.Serial(port=port2try, baudrate=115200, parity=serial.PARITY_NONE, timeout=.1)
            time.sleep(.2)
            args = activate_gateway(args, ser)
            break
        except IndexError:
            if ser is not None:
                ser.close()
                ser = None
            continue

    if ser is None:
        log.error('No USB Sensor Adapter Detected!')
        raise serial.SerialException('No USB Sensor Adapter Detected!')
    else:
        log.info('Serial port: {}'.format(ser.port))
    return ser


def start_usb_pro():
    log = AppLogger()
    global arguments
    global error_retry_count
    pro_thread = threading.currentThread()
    parser = get_parser()
    arguments = parser.parse_args()
    config = ConfigParser()
    filecheck = os.path.isfile('config.ini')
    lock_config_file('start_usb_pro()')
    if filecheck:
        config.read('config.ini')
    else:
        log.info('Could not find config.ini, rebuilding.')
        rebuild(config)
        config.read('config.ini')
    unlock_config_file('start_usb_pro()')
    arguments.heartbeat = int(config.get('usb_configs', 'heartbeat'))
    arguments.sleep_time = int(config.get('usb_configs', 'sleep_time'))
    arguments.encryption = config.get('usb_configs', 'encryption')
    arguments.host = constants.MONNIT_HOST
    arguments.pretty = True
    # arguments.host = config.get('usb_configs', 'host')
    constants.BLFlag = False
    if constants.VERBOSE_LOGS:
        log.info('Heartbeat in minutes: {}'.format(arguments.heartbeat))
        log.info('Upqueue request time in seconds: {}'.format(arguments.sleep_time))
        log.info('Encryption: {}'.format(arguments.encryption))
    PayloadMessage.args = arguments
    error_retry_count = 0
    # Run the main loop while handling exceptions
    ser = None
    while (error_retry_count < int(arguments.max_error_retries) or arguments.no_retry_limit) & (getattr(pro_thread, "do_run", True)):
        try:
            ser = serial_search(arguments, ser)
            run_usb_pro(arguments, ser)
        except ValueError as ve:
            if not arguments.quiet:
                log.error('Argument error: {}'.format(USBException(str(ve)).as_json()))
        except KeyboardInterrupt:
            log.error('Forced exit due to keyboard interrupt.')
            sys.exit()
        except serial.SerialException:
            usb_exception = USBException('Cannot connect to serial device: [{}] Retrying in {} seconds...'.format(arguments.input, arguments.error_sleep_time)).as_json()
            log.error('Serial Exception: {}'.format(usb_exception))
            time.sleep(int(arguments.error_sleep_time))
            error_retry_count += 1
        except Exception as e:
            if not arguments.quiet:
                usb_exception = USBException(str(e)).as_json()
                log.error('Unknown Exception: {}'.format(usb_exception))
                log.error('Traceback: {}'.format(traceback.format_exc()))
        if ser is not None:
            try:
                ser.close()
            except Exception as fail:
                log.info('Unable to close serial port {}'.format(fail))

        error_retry_count += 1

        time.sleep(1)
