
__all__ = ["make_keypair", "make_publickey", "compute_sharedsecret", "compute_aes_keyiv"]

import collections
import random
# import struct
# import array

# Elliptic Curve defined as Y^2 = X^3 + aX + b
EllipticCurve = collections.namedtuple('EllipticCurve', 'name s p a b g n h')
curve = EllipticCurve(
    name='secp256r1',
    s=0xc49d360886e704936a6678e1139d26b7819f7e90,
    # Field characteristic.
    p=0xffffffff00000001000000000000000000000000ffffffffffffffffffffffff,
    # Curve coefficients.
    a=0xffffffff00000001000000000000000000000000fffffffffffffffffffffffc,
    b=0x5ac635d8aa3a93e7b3ebbd55769886bc651d06b0cc53b0f63bce3c3e27d2604b,
    # Base point.
    g=(0x6b17d1f2e12c4247f8bce6e563a440f277037d812deb33a0f4a13945d898c296,
       0x4fe342e2fe1a7f9b8ee7eb4a7c0f9e162bce33576b315ececbb6406837bf51f5),
    # Subgroup order.
    n=0xffffffff00000000ffffffffffffffffbce6faada7179e84f3b9cac2cf632551,
    # Subgroup cofactor.
    h=1,
)


# Modular arithmetic ##########################################################

def inverse_mod(k, p):
    """Returns the inverse of k modulo p.

    This function returns the only integer x such that (x * k) % p == 1.

    k must be non-zero and p must be a prime.
    """
    if k == 0:
        raise ZeroDivisionError('division by zero')

    if k < 0:
        # k ** -1 = p - (-k) ** -1  (mod p)
        return p - inverse_mod(-k, p)

    # Extended Euclidean algorithm.
    s, old_s = 0, 1
    t, old_t = 1, 0
    r, old_r = p, k

    while r != 0:
        quotient = old_r // r
        old_r, r = r, old_r - quotient * r
        old_s, s = s, old_s - quotient * s
        old_t, t = t, old_t - quotient * t

    gcd, x, y = old_r, old_s, old_t

    assert gcd == 1
    assert (k * x) % p == 1

    return x % p


# Functions that work on curve points #########################################
# Elliptic Curve defined as Y^2 = X^3 + aX + b
def is_on_curve(point):
    """Returns True if the given point lies on the elliptic curve."""
    if point is None:  # None represents the point at infinity.
        return True
    x, y = point
    return (y * y - x * x * x - curve.a * x - curve.b) % curve.p == 0


def isqrt(x):
    if x < 0:
        raise ValueError('square root not defined for negative numbers')
    n = int(x)
    if n == 0:
        return 0
    a, b = divmod(n.bit_length(), 2)
    x = 2**(a+b)
    while True:
        y = (x + n//x)//2
        if y >= x:
            return x
        x = y


def point_neg(point):
    """Returns -point."""
    assert is_on_curve(point)

    if point is None:
        # -0 = 0
        return None

    x, y = point
    result = (x, -y % curve.p)

    assert is_on_curve(result)

    return result


def point_add(point1, point2):
    """Returns the result of point1 + point2 according to the group law."""
    assert is_on_curve(point1)
    assert is_on_curve(point2)

    if point1 is None:
        # 0 + point2 = point2
        return point2
    if point2 is None:
        # point1 + 0 = point1
        return point1

    x1, y1 = point1
    x2, y2 = point2

    if x1 == x2 and y1 != y2:
        # point1 + (-point1) = 0
        return None

    if x1 == x2:
        # This is the case point1 == point2.
        m = (3 * x1 * x1 + curve.a) * inverse_mod(2 * y1, curve.p)
    else:
        # This is the case point1 != point2.
        m = (y1 - y2) * inverse_mod(x1 - x2, curve.p)

    x3 = m * m - x1 - x2
    y3 = y1 + m * (x3 - x1)
    result = (x3 % curve.p,
              -y3 % curve.p)

    assert is_on_curve(result)

    return result


def scalar_mult(k, point):
    """Returns k * point computed using the double and point_add algorithm."""
    assert is_on_curve(point)

    if k % curve.n == 0 or point is None:
        return None

    if k < 0:
        # k * point = -k * (-point)
        return scalar_mult(-k, point_neg(point))

    result = None
    addend = point

    while k:
        if k & 1:
            # Add.
            result = point_add(result, addend)

        # Double.
        addend = point_add(addend, addend)

        k >>= 1

    assert is_on_curve(result)

    return result


def make_keypair():
    """Generates a random private-public key pair."""
    private_key = random.randrange(1, curve.n)
    public_key = scalar_mult(private_key, curve.g)

    return private_key, public_key


def make_publickey(private_key):
    """Generates a specific private-public key pair."""
    public_key = scalar_mult(private_key, curve.g)
    return public_key


def compute_sharedsecret(private_key, other_public_key):
    """Compute a shared secret for ECDH."""
    ss = scalar_mult(private_key, other_public_key)
    return ss


def compute_aes_keyiv(ss, ver):
    """Take a shared secret and convert to AES Key/IV based on ver input"""
    a = list(ss[0].to_bytes(32, byteorder='little', signed=False))
    a += list(ss[1].to_bytes(32, byteorder='little', signed=False))
    key = None
    iv = None
    if ver == 0:
        # KEY 12 62 31 29 19 1  56 5  39 58 16 21 4  36 63 32
        # IV  48 8  3  37 18 24 61 17 54 13 44 22 23 46 7  17
        key = [a[12], a[62], a[31], a[29], a[19], a[1], a[56], a[5],
               a[39], a[58], a[16], a[21], a[4], a[36], a[63], a[32]]
        iv = [a[48], a[8], a[3], a[37], a[18], a[24], a[61], a[17],
              a[54], a[13], a[44], a[22], a[23], a[46], a[7], a[17]]
    return key, iv



